"""
Execution plan entity.
"""
import os
import textwrap
import time
import uuid

import yaml
from yamlable import YamlAble, yaml_info

from vmm_manager.entity.action import Action
from vmm_manager.infra.command import Command
from vmm_manager.scvmm.scjob import SCJob
from vmm_manager.util.msgs import (formatar_msg_erro, imprimir_acao_corrente,
                                   imprimir_erro, imprimir_ok)


@yaml_info(yaml_tag_ns='scvmm_manager')
class Plan(YamlAble):
    ARQUIVO_PLANO_EXECUCAO = 'plan.yaml'
    ARQUIVO_LOG_ERROS = 'errors.log'

    @staticmethod
    def carregar_plano_execucao(execution_plan_file):
        try:
            with open(execution_plan_file, 'r', encoding='utf8') as file:
                plano_execucao = yaml.safe_load(file)
                return True, plano_execucao
        except (IOError, TypeError) as erro:
            return False, f"Error loading the plan '{execution_plan_file}'.\n{erro}"

    @staticmethod
    def excluir_arquivo():
        if os.path.exists(Plan.ARQUIVO_PLANO_EXECUCAO):
            os.remove(Plan.ARQUIVO_PLANO_EXECUCAO)

    @staticmethod
    def __excluir_arquivo_log_erros():
        if os.path.exists(Plan.ARQUIVO_LOG_ERROS):
            os.remove(Plan.ARQUIVO_LOG_ERROS)

    def __init__(self, group, cloud):
        self.group = group
        self.cloud = cloud
        self.actions = []
        self.interval_between_resources = 0

        self.__jobs_em_execucao = {}
        self.__guids_a_limpar = []
        self.__msgs_erros = ''

    def is_vazio(self):
        return not self.actions

    def gerar_arquivo(self):
        try:
            conteudo = yaml.safe_dump(self, default_flow_style=False)
            with open(Plan.ARQUIVO_PLANO_EXECUCAO, 'w', encoding='utf8') as arquivo_yaml:
                arquivo_yaml.write(conteudo)
        except IOError as erro:
            return False, f'Error generating file {Plan.ARQUIVO_PLANO_EXECUCAO}.\n{erro}'

        return True, conteudo

    def __executar_acoes(self, actions, servidor_acesso, ocultar_progresso):
        if not actions:
            return  # return early if there are no actions to execute

        # zerando jobs em execução
        self.__jobs_em_execucao = {}

        total_acoes = len(actions)
        for i in range(total_acoes):
            acao = actions[i]

            print(f'{textwrap.shorten(acao.get_str_impressao_inline(), 100)} => ',
                  end='', flush=True)

            # Caso específico de criação de vms
            guid = uuid.uuid4()
            if acao.is_criacao_vm():
                self.__guids_a_limpar.append(guid)

            acao.executar(self.group, self.cloud,
                          servidor_acesso, guid)

            # Tratando resultado
            if acao.was_executada_com_sucesso():
                guid = acao.get_resultado_execucao().get('Guid')
                if guid:
                    # Job em execução
                    self.__jobs_em_execucao[guid] = SCJob(
                        guid, acao)
                    print(f'[Started {guid}]')
                else:
                    # Comando já finalizado
                    imprimir_ok(ocultar_progresso)
            else:
                imprimir_erro(ocultar_progresso)
                self.__logar_erros_acao(
                    acao, acao.get_resultado_execucao().get('Msgs'))

            # If the action is not the last, check if we need to wait
            if i < total_acoes - 1:
                proxima_acao = actions[i + 1]

                if not acao.is_same_resource(proxima_acao) and self.interval_between_resources > 0:
                    print(
                        f'{textwrap.shorten(f"Waiting {self.interval_between_resources} seconds to start the next action", 100)} => ',  # noqa: E501
                        end='',
                        flush=True
                    )
                    time.sleep(self.interval_between_resources)
                    imprimir_ok(ocultar_progresso)

        # Monitorando jobs
        SCJob.monitore_jobs(self.__jobs_em_execucao, servidor_acesso)
        self.__coletar_resultado_jobs()

        # Ações pós execução
        print()
        self.__executar_cmds_finalizacao(
            actions, servidor_acesso, ocultar_progresso)

    def executar(self, servidor_acesso, ocultar_progresso, interval_between_resources=0):
        if self.is_vazio():
            return

        # update interval between resources
        if interval_between_resources >= 0:
            self.interval_between_resources = interval_between_resources

        print('\nApplied operations:')

        # actions bloqueantes primeiro
        acoes_bloqueantes = [
            acao for acao in self.actions if acao.is_bloqueante()]
        self.__executar_acoes(
            acoes_bloqueantes, servidor_acesso, ocultar_progresso)

        # demais ações, se não houve erro
        if not self.has_erro_execucao():
            acoes_nao_bloqueantes = [acao
                                     for acao in self.actions if not acao.is_bloqueante()]
            self.__executar_acoes(
                acoes_nao_bloqueantes, servidor_acesso, ocultar_progresso)

        # Ações de finalização
        self.__limpar_guids(servidor_acesso, ocultar_progresso)
        self.__processa_resultado_execucao()
        Plan.excluir_arquivo()

    def __executar_cmds_finalizacao(self, actions, servidor_acesso, ocultar_progresso):
        for acao in actions:
            if acao.was_executada_com_sucesso() and acao.has_cmd_pos_execucao():
                cmd = acao.get_cmd_pos_execucao(
                    self.group, servidor_acesso)

                imprimir_acao_corrente(cmd.description, ocultar_progresso)

                status, resultado = cmd.executar(servidor_acesso)

                if status:
                    imprimir_ok(ocultar_progresso)
                else:
                    imprimir_erro(ocultar_progresso)
                    self.__logar_erros_comando(cmd.description, resultado)

    def __limpar_guids(self, servidor_acesso, ocultar_progresso):
        if self.__guids_a_limpar:
            imprimir_acao_corrente(
                'Cleaning up temporary VMM objects', ocultar_progresso)

            cmd = Command(
                'clean_objects_after_vm_creation',
                vmm_server=servidor_acesso.vmm_server,
                guids=self.__guids_a_limpar
            )
            status, retorno = cmd.executar(servidor_acesso)

            if status:
                imprimir_ok(ocultar_progresso)
            else:
                imprimir_erro(ocultar_progresso)
                self.__logar_erros_comando(
                    'Clean up temporary VMM objects', retorno)

    def __processa_resultado_execucao(self):
        if self.has_erro_execucao():
            self.__gerar_arquivo_erros()
        else:
            Plan.__excluir_arquivo_log_erros()

    def imprimir_resultado_execucao(self):
        if self.has_erro_execucao():
            print(formatar_msg_erro(
                '\nSynchronization completed with errors. '
                f'More details in {Plan.ARQUIVO_LOG_ERROS}.'))
        else:
            print('\nSynchronization completed successfully.')

    def has_erro_execucao(self):
        return self.__msgs_erros

    def __coletar_resultado_jobs(self):
        if self.__jobs_em_execucao:
            for job in self.__jobs_em_execucao.values():
                if job.is_finalizado_com_erro():
                    print(formatar_msg_erro(
                        f'Process {job.vmm_id} finalized with error.'))
                    self.__logar_erros_acao(job.acao, job.resumo_erro)

    def __logar_erros_comando(self, command, erro):
        self.__msgs_erros += f'Error found in "{command}":\n{erro}\n\n'

    def __logar_erros_acao(self, acao, erro):
        self.__msgs_erros += f'Command:\n{acao.get_str_impressao_inline()}' \
            f'\n\nCaptured error:\n{erro}\n\n'

    def __gerar_arquivo_erros(self):
        try:
            with open(Plan.ARQUIVO_LOG_ERROS, 'w', encoding='utf8') as arquivo_erros:
                arquivo_erros.write(self.__msgs_erros)
        except IOError as erro:
            print(
                f'It was not possible to generate the error file ({Plan.ARQUIVO_LOG_ERROS}): '
                f'{erro}')

    def imprimir_acoes(self):
        for acao in self.actions:
            print(acao.get_str_impressao_inline())

    def __eq__(self, other):
        return isinstance(other, Plan) and (self.group == other.group
                                            and self.cloud == other.cloud
                                            and self.actions == other.actions)

    def __str__(self):
        return f'''
            group: {self.group}
            cloud: {self.cloud}
            actions: {','.join(str(acao) for acao in self.actions)}
            '''

    def __to_yaml_dict__(self):
        return {'group': self.group,
                'cloud': self.cloud,
                'actions': self.actions}

    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag):
        plano_execucao = Plan(dct['group'], dct['cloud'])

        for acao_str in dct['actions']:
            acao = Action(acao_str.command, **acao_str.args['args'])
            plano_execucao.actions.append(acao)

        return plano_execucao



"""
Módulo relacionado à preparação de comandos para execução no servidor de acesso
"""
import os
import sys

from jinja2 import Environment, FileSystemLoader, exceptions


class Command:
    __TEMPLATES_DIR = '../includes/ps_templates'
    __TEMPLATES_EXTENSION = '.j2'

    def __init__(self, command, description=None, **kwargs):
        try:
            self.command = command
            self.description = description if description else self.command
            self.args = kwargs
            j2_env = Environment(
                loader=FileSystemLoader(os.path.join(
                    os.path.dirname(__file__), Command.__TEMPLATES_DIR)),
                trim_blocks=True)
            self.template = j2_env.get_template(
                self.command + Command.__TEMPLATES_EXTENSION)
        except exceptions.TemplateNotFound as ex:
            print(
                f"Template '{ex}' not found for the command '{self.command}'.")
            sys.exit(1)

    def imprimir(self):
        print('\n' + self.template.render(self.args) + '\n')

    def executar(self, servidor_acesso):
        conteudo_script = self.template.render(self.args)
        return servidor_acesso.executar_script(
            self.command, conteudo_script)



"""
VM additional disk entity.
"""
from vmm_manager.entity.action import Action


class VMDisk:
    def __init__(self, bus_type, file, size_mb, size_type, path=None):
        self.bus_type = bus_type
        self.file = file
        self.size_mb = size_mb
        self.size_type = size_type
        self.path = path

        self.__id_drive = None
        self.__id_disco = None
        self.__bus = None
        self.__lun = None

    def set_parametros_extras_vmm(self, drive_id, disk_id, bus, lun):
        self.__id_drive = drive_id
        self.__id_disco = disk_id
        self.__bus = bus
        self.__lun = lun

    def get_id_disco(self):
        if not self.__id_disco:
            raise ValueError('Local disk does not have an ID.')

        return self.__id_disco

    def get_id_drive(self):
        if not self.__id_drive:
            raise ValueError('Local drive does not have an ID.')

        return self.__id_drive

    def get_tamanho_tipo_create(self):
        return self.size_type.name

    def get_acao_criar_disco(self, vm_name):
        return Action(
            'create_vm_disk',
            vm_name=vm_name,
            bus_type=self.bus_type.value,
            size_mb=self.size_mb,
            size_type=self.get_tamanho_tipo_create(),
            file=self.file,
            path=self.path
        )

    def get_acao_excluir_disco(self, vm_id):
        return Action(
            Action.ACAO_EXCLUIR_DISCO_VM,
            vm_id=vm_id,
            drive_id=self.get_id_drive()
        )

    def get_acao_expandir_disco(self, vm_id, drive_id):
        return Action(
            'expand_vm_disk',
            vm_id=vm_id,
            drive_id=drive_id,
            size_mb=self.size_mb
        )

    def get_acao_mover_disco(self, vm_id, disk_id):
        return Action(
            'mover_disco_vm',
            vm_id=vm_id,
            disk_id=disk_id,
            path=self.path
        )

    def get_acao_converter_disco(self, vm_id, drive_id):
        return Action(
            'convert_vm_disk',
            vm_id=vm_id,
            drive_id=drive_id,
            size_type=self.get_tamanho_tipo_create()
        )

    def get_acoes_diferenca_disco(self, disco_remoto, vm_id, vm_name):
        actions = []

        # Alteração de tipo ou redução de disco exige a recriação
        if ((self.bus_type != disco_remoto.bus_type) or
                (self.size_mb < disco_remoto.size_mb)):
            actions.append(disco_remoto.get_acao_excluir_disco(vm_id))
            actions.append(self.get_acao_criar_disco(vm_name))
        else:
            # Type de tamanho alterado: converter disco
            if self.size_type != disco_remoto.size_type:
                actions.append(self.get_acao_converter_disco(
                    vm_id, disco_remoto.get_id_drive()))

            # Tamanho alterado: expansão
            if self.size_mb > disco_remoto.size_mb:
                actions.append(self.get_acao_expandir_disco(
                    vm_id, disco_remoto.get_id_drive()))

            # Path alterado: mover disco
            if self.path and self.path != disco_remoto.path:
                actions.append(self.get_acao_mover_disco(
                    vm_id, disco_remoto.get_id_disco()))

        return actions

    def __hash__(self):
        return hash(self.file)

    def __eq__(self, other):
        return isinstance(other, VMDisk) and (self.bus_type == other.bus_type
                                              and self.file == other.file
                                              and self.size_mb == other.size_mb
                                              and self.size_type == other.size_type
                                              and self.path == other.path)

    def __repr__(self):
        return f'''
                bus_type: {self.bus_type}
                file: {self.file}
                size_mb: {self.size_mb}
                size_type: {self.size_type}
                path: {self.path}
                drive_id: {self.__id_drive}
                disk_id: {self.__id_disco}
                bus: {self.__bus}
                lun: {self.__lun}
                '''

    def to_dict(self):
        return {
            'bus_type': self.bus_type.value,
            'file': self.file,
            'size_mb': self.size_mb,
            'size_type': self.size_type.value,
            'path': self.path,
            'drive_id': self.__id_drive,
            'disk_id': self.__id_disco,
            'bus': self.__bus,
            'lun': self.__lun
        }



"""
VM Ansible entity.
"""


class VMAnsible:
    def __init__(self, group):
        self.group = group
        self.variaveis = []

    def extrair_dados_vars_dict(self, dict_vars, vm_name):
        lista_nomes_var = []
        for item in dict_vars or {}:
            nome_var = item.get('name')

            if nome_var in lista_nomes_var:
                raise ValueError(
                    (f"Variable '{nome_var}' from Ansible group '{self.group}' "
                     f"already exists in VM '{vm_name}'."))

            lista_nomes_var.append(nome_var)
            ansible_var = VMAnsibleVars(
                nome_var, item.get('value'))

            self.variaveis.append(ansible_var)

    def __hash__(self):
        return hash(self.group)

    def __eq__(self, other):
        return isinstance(other, VMAnsible) and (self.group == other.group
                                                 and self.variaveis == other.variaveis)

    def __repr__(self):
        return f'''
                group: {self.group}
                vars: {self.variaveis}
                '''

    def to_dict(self):
        return {
            'group': self.group,
            'vars': [variavel.to_dict() for variavel in self.variaveis]
        }


class VMAnsibleVars:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __eq__(self, other):
        return isinstance(other, VMAnsibleVars) and (self.name == other.name
                                                     and self.value == other.value)

    def __repr__(self):
        return f'''
                name: {self.name}
                value: {self.value}
                '''

    def to_dict(self):
        return {
            'name': self.name,
            'value': self.value
        }



"""
System Center Virtual Machine Manager (SCVMM) job.
"""

import json
from time import sleep

import tqdm

from vmm_manager.infra.command import Command
from vmm_manager.scvmm.enums import SCJobStatusEnum
from vmm_manager.util.msgs import formatar_msg_aviso


class SCJob():

    @staticmethod
    def monitore_jobs(jobs, servidor_acesso):
        if jobs:
            print('\nMonitoring jobs:')
            pbars = {}
            for job in jobs.values():
                pbar = tqdm.tqdm(total=100)
                pbar.set_description(job.vmm_id)
                pbars[job.vmm_id] = pbar

            # Monitorando jobs do VMM
            tem_jobs_em_andamento = len(pbars) > 0
            while tem_jobs_em_andamento:
                tem_jobs_em_andamento = False
                status, jobs_vmm = SCJob.__obter_status_jobs_vmm(
                    jobs.keys(), servidor_acesso)

                # No caso de erro na api, finalizar loop
                if not status:
                    tqdm.tqdm.write(
                        formatar_msg_aviso(
                            '\nError while monitoring jobs.\n'))
                    break

                # Atualizando status
                for job_vmm in json.loads(jobs_vmm):
                    vmm_id = job_vmm.get('ID')
                    jobs[vmm_id].atualizar(job_vmm)
                    pbars[vmm_id].update(jobs[vmm_id].ultimo_progresso)
                    pbars[vmm_id].refresh()

                    if not jobs[vmm_id].is_finalizado:
                        tem_jobs_em_andamento = True

                # Aguardando um tempo para próxima checagem
                if tem_jobs_em_andamento:
                    sleep(10)

            # finalizando barra
            for pbar in pbars.values():
                pbar.close()

    @staticmethod
    def __obter_status_jobs_vmm(jobs, servidor_acesso):
        cmd = Command('monitore_jobs',
                      vmm_server=servidor_acesso.vmm_server,
                      jobs=jobs)
        return cmd.executar(servidor_acesso)

    def __init__(self, vmm_id, acao):
        self.vmm_id = vmm_id
        self.acao = acao

        self.name = None
        self.status = None
        self.is_finalizado = None
        self.resumo_erro = None

        self.__is_sucesso = None
        self.__codigo_erro = None
        self.__msg_erro = None
        self.__acao_recomendada = None

        self.__progresso_anterior = 0
        self.progresso = 0
        self.ultimo_progresso = 0

        self.__processar_finalizacao_job()

    def is_finalizado_com_erro(self):
        return self.is_finalizado and not self.__is_sucesso

    def __processar_finalizacao_job(self):
        if self.is_finalizado:
            self.acao.informar_status_execucao_job(
                not self.is_finalizado_com_erro())

    def atualizar(self, job_vmm):
        self.name = job_vmm.get('Name')
        self.status = SCJobStatusEnum(job_vmm.get('Status'))
        self.is_finalizado = job_vmm.get('IsCompleted')

        self.__progresso_anterior = self.progresso
        self.progresso = job_vmm.get('ProgressValue')
        self.ultimo_progresso = self.progresso - self.__progresso_anterior

        if job_vmm.get('ErrorInfo'):
            self.__is_sucesso = job_vmm.get('ErrorInfo').get('IsSuccess')
            self.__codigo_erro = job_vmm.get('ErrorInfo').get('Code')
            self.__msg_erro = job_vmm.get('ErrorInfo').get('Problem')
            self.__acao_recomendada = job_vmm.get(
                'ErrorInfo').get('RecommendedAction')
            self.resumo_erro = job_vmm.get(
                'ErrorInfo').get('CSMMessageString')

        self.__processar_finalizacao_job()

    def __str__(self):
        return f'''
            vmm_id: {self.vmm_id}
            acao: {self.acao}
            name: {self.name}
            status: {self.status}
            is_finalizado: {self.is_finalizado}
            progresso: {self.progresso}
            is_sucesso: {self.__is_sucesso}
            codigo_erro: {self.__codigo_erro}
            msg_erro: {self.__msg_erro}
            acao_recomendada: {self.__acao_recomendada}
            resumo_erro: {self.resumo_erro}
            '''



"""
System Center Virtual Machine Manager (SCVMM) region.
"""

# pylint: disable=too-few-public-methods


class SCRegion():
    REGION_DEFAULT = 'default'

    def __init__(self, id_no, nome_no, group, cluster, letra_id=None):
        self.id_no = id_no
        self.nome_no = nome_no
        self.group = group
        self.cluster = cluster
        self.letra_id = letra_id

    def __str__(self):
        return f'''
            id_no: {self.id_no}
            nome_no: {self.nome_no}
            group: {self.group}
            cluster: {self.cluster}
            letra_id: {self.letra_id}
            '''



"""
Módulo que realiza o parser de um inventário remoto (no SCVMM)
"""
import json

from vmm_manager.entity.inventory import Inventory
from vmm_manager.entity.vm import VM
from vmm_manager.entity.vm_disk import VMDisk
from vmm_manager.entity.vm_network import VMNetwork
from vmm_manager.infra.command import Command
from vmm_manager.scvmm.enums import SCDiskBusType, SCDiskSizeType, VMStatusEnum
from vmm_manager.scvmm.scregion import SCRegion
from vmm_manager.util.config import (FIELD_GROUP, FIELD_ID, FIELD_IMAGE,
                                     FIELD_NETWORK_DEFAULT, FIELD_REGION)


# pylint: disable=too-few-public-methods
class ParserRemote:

    @staticmethod
    def __get_regioes_disponiveis(servidor_acesso):
        cmd = Command('get_available_regions',
                      vmm_server=servidor_acesso.vmm_server)
        status, regioes = cmd.executar(servidor_acesso)
        if not status:
            raise Exception(  # pylint: disable=broad-exception-raised
                f'Error getting available regions: {regioes}')

        regioes_remoto = json.loads(regioes)
        regioes_disponiveis = []
        for region in regioes_remoto:
            regiao_obj = SCRegion(
                region.get('HostID'),
                region.get('Hostname'),
                region.get('Group'),
                region.get('Cluster')
            )

            regioes_disponiveis.append(regiao_obj)

        return regioes_disponiveis

    def __init__(self, group, cloud):
        self.group = group
        self.cloud = cloud
        self.__inventario = None

    def __get_vms_servidor(self, servidor_acesso, filtro_nome_vm=None):
        cmd = Command('get_vms_in_group',
                      vmm_server=servidor_acesso.vmm_server,
                      field_group=FIELD_GROUP[0],
                      field_id=FIELD_ID[0],
                      field_image=FIELD_IMAGE[0],
                      field_region=FIELD_REGION[0],
                      field_network_default=FIELD_NETWORK_DEFAULT[0],
                      group=self.group,
                      filtro_nome_vm=filtro_nome_vm,
                      cloud=self.cloud)
        status, vms = cmd.executar(servidor_acesso)
        if not status:
            raise Exception(  # pylint: disable=broad-exception-raised
                f"Error getting VMs: {vms}")
        return vms

    def __get_discos_adicionais(self, servidor_acesso):
        cmd = Command('get_additional_disks',
                      vmm_server=servidor_acesso.vmm_server,
                      field_group=FIELD_GROUP[0],
                      field_id=FIELD_ID[0],
                      group=self.group,
                      cloud=self.cloud,
                      vm_nomes=','.join([f'"{vm_name}"' for vm_name in self.__inventario.vms]))
        status, additional_disks = cmd.executar(servidor_acesso)
        if not status:
            raise Exception(  # pylint: disable=broad-exception-raised
                f'Error getting additional disks: {additional_disks}')

        discos_vms_remoto = json.loads(additional_disks)
        discos_vms = {}
        for maquina_virtual in discos_vms_remoto:
            vm_name = maquina_virtual.get('Name')
            discos_vms[vm_name] = []

            for disco_remoto in maquina_virtual.get('Discos'):
                disco = VMDisk(
                    SCDiskBusType(disco_remoto.get('Type')),
                    disco_remoto.get('File'),
                    disco_remoto.get('SizeMB'),
                    SCDiskSizeType(disco_remoto.get('SizeType')),
                    disco_remoto.get('Path'))

                disco.set_parametros_extras_vmm(
                    disco_remoto.get('DriveID'),
                    disco_remoto.get('DiskID'),
                    disco_remoto.get('Bus'),
                    disco_remoto.get('Lun'),
                )

                discos_vms[vm_name].append(disco)

        return discos_vms

    def __montar_inventario(
        self,
        servidor_acesso,
        filtro_nome_vm=None,
        filtro_dados_completos=True
    ):
        self.__inventario = Inventory(self.group, self.cloud)

        vms_servidor = json.loads(
            self.__get_vms_servidor(servidor_acesso, filtro_nome_vm) or '{}')

        for maquina_virtual in vms_servidor:
            vms_rede = []
            for network in maquina_virtual.get('Networks'):
                vm_rede = VMNetwork(network.get('Name'), network.get('Principal'))
                vm_rede.ips = network.get('IPS', '').split(' ')
                vms_rede.append(vm_rede)

            self.__inventario.vms[maquina_virtual.get('Name')] = VM(
                maquina_virtual.get('Name'),
                maquina_virtual.get('Description'),
                maquina_virtual.get('Image'),
                maquina_virtual.get('Region'),
                maquina_virtual.get('Cpu'),
                maquina_virtual.get('Ram'),
                vms_rede,
                maquina_virtual.get('ID'),
                maquina_virtual.get('NestedVirtualization'),
                maquina_virtual.get('DynamicMemory'),
                VMStatusEnum(maquina_virtual.get('Status')),
                maquina_virtual.get('RegionHostname'),
            )

        # Obtendo dados adicionais
        if filtro_dados_completos:
            # discos
            self.__inventario.set_discos_vms(
                self.__get_discos_adicionais(servidor_acesso))

            # regioes
            self.__inventario.set_regioes_disponiveis(
                ParserRemote.__get_regioes_disponiveis(servidor_acesso))

    def get_inventario(self, servidor_acesso, filtro_nome_vm=None, filtro_dados_completos=True):
        if not self.__inventario:
            try:
                self.__montar_inventario(
                    servidor_acesso, filtro_nome_vm, filtro_dados_completos)
            # pylint: disable=broad-except
            except Exception as ex:
                return False, str(ex)

        return True, self.__inventario



"""
VM entity.
"""
from vmm_manager.entity.action import Action
from vmm_manager.entity.vm_ansible import VMAnsible
from vmm_manager.scvmm.enums import VMStatusEnum
from vmm_manager.scvmm.scregion import SCRegion


class VM:
    def __init__(
        self,
        name,
        description,
        image,
        region,
        cpu,
        memory,
        networks,
        vmm_id=None,
        nested_virtualization=False,
        dynamic_memory=True,
        status=VMStatusEnum.RUNNING,
        region_host=None
    ):
        self.name = name
        self.description = description if description is not None else ''
        self.image = image
        self.region = region
        self.cpu = cpu
        self.memory = memory
        self.networks = networks
        self.vmm_id = vmm_id
        self.status = status
        self.region_host = region_host
        self.nested_virtualization = nested_virtualization
        self.dynamic_memory = dynamic_memory

        self.dados_ansible = {}
        self.additional_disks = {}
        self.to_json_dados_completos = True

    def extrair_dados_ansible_dict(self, dict_ansible):
        for item in dict_ansible or {}:
            group = item.get('group')

            if group in self.dados_ansible:
                raise ValueError(
                    f"Ansible group '{group}' "
                    f"already exists in VM '{self.name}'.")

            ansible_grupo = VMAnsible(group)
            ansible_grupo.extrair_dados_vars_dict(
                item.get('vars'), self.name)

            self.dados_ansible[group] = ansible_grupo

    def add_discos_adicionais(self, additional_disks):
        for additional_disk in additional_disks:
            self.additional_disks[additional_disk.file] = additional_disk

    def add_acoes_diferenca_discos_adicionais(self, vm_remota, plano_execucao):
        # discos a excluir
        if vm_remota:
            discos_excluir = [
                nome_disco_remoto for nome_disco_remoto in vm_remota.additional_disks
                if nome_disco_remoto not in self.additional_disks
            ]
            for nome_disco in discos_excluir:
                plano_execucao.actions.append(
                    vm_remota.additional_disks[nome_disco].get_acao_excluir_disco(
                        vm_remota.vmm_id))

        # verificando discos atuais
        for nome_disco, data_disco in self.additional_disks.items():
            # discos a criar
            if not vm_remota or nome_disco not in vm_remota.additional_disks:
                plano_execucao.actions.append(
                    data_disco.get_acao_criar_disco(self.name))
            else:
                # discos a alterar
                plano_execucao.actions.extend(
                    data_disco.get_acoes_diferenca_disco(
                        vm_remota.additional_disks[nome_disco], vm_remota.vmm_id, self.name))

    def add_acoes_diferenca_regiao(
        self,
        vm_remota,
        plano_execucao,
        inv_remoto
    ):
        if self.region == SCRegion.REGION_DEFAULT:
            return

        if (not vm_remota
            or (self.region != vm_remota.region or
                inv_remoto.get_nome_no_regiao(self.region) != vm_remota.region_host)):
            plano_execucao.actions.append(self.get_acao_mover_vm_regiao(
                inv_remoto.get_id_no_regiao(self.region)))

    def add_acoes_virtualizacao_aninhada(self, vm_remota,
                                         plano_execucao):
        if ((not vm_remota and self.nested_virtualization)
                or (vm_remota and vm_remota.nested_virtualization != self.nested_virtualization)):
            plano_execucao.actions.append(
                self.get_acao_atualizar_virtualizacao_aninhada())

    def add_acoes_memoria_dinamica(self, vm_remota,
                                   plano_execucao):
        if (vm_remota and vm_remota.dynamic_memory != self.dynamic_memory):
            plano_execucao.actions.append(
                self.get_acao_atualizar_memoria_dinamica())

    def add_acoes_diferenca_vm(self, vm_remota, plano_execucao):
        # alteração da image é irreversível
        if self.image != vm_remota.image:
            plano_execucao.actions.append(vm_remota.get_acao_excluir_vm())
            plano_execucao.actions.append(self.get_acao_criar_vm())
            return

        # Alteração de network é possível recuperar TODO #18
        if self.networks != vm_remota.networks:
            plano_execucao.actions.append(vm_remota.get_acao_excluir_vm())
            plano_execucao.actions.append(self.get_acao_criar_vm())
            return

        # alteração de descrição, cpu ou ram
        if (self.description != vm_remota.description
                or self.cpu != vm_remota.cpu
                or self.memory != vm_remota.memory):
            plano_execucao.actions.append(
                self.get_acao_atualizar_vm(vm_remota.vmm_id))

    def get_qtde_rede_principal(self):
        return sum(1 for network in self.networks if network.default)

    def get_rede_principal(self):
        return next((network.name for network in self.networks if network.default), None)

    def get_acao_criar_vm(self):
        return Action(
            Action.ACAO_CRIAR_VM,
            vm_name=self.name,
            description=self.description,
            image=self.image,
            region=self.region,
            cpu=self.cpu,
            memory=self.memory,
            dynamic_memory=self.dynamic_memory,
            networks=[network.name for network in self.networks],
            network_default=self.get_rede_principal()
        )

    def get_acao_excluir_vm(self):
        return Action(
            Action.ACAO_EXCLUIR_VM,
            vm_id=self.vmm_id
        )

    def get_acao_mover_vm_regiao(self, region_host_id):
        return Action(
            'move_vm_region',
            vm_name=self.name,
            region_host_id=region_host_id,
            region=self.region
        )

    def get_acao_atualizar_virtualizacao_aninhada(self):
        return Action(
            'update_nested_virtualization',
            vm_name=self.name,
            nested_virtualization=self.nested_virtualization
        )

    def get_acao_atualizar_memoria_dinamica(self):
        return Action(
            'update_dynamic_memory',
            vm_name=self.name,
            dynamic_memory=self.dynamic_memory
        )

    def get_acao_atualizar_vm(self, id_vm_remota):
        return Action(
            'update_vm',
            vm_id=id_vm_remota,
            description=self.description,
            cpu=self.cpu,
            memory=self.memory
        )

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, VM) and (self.name == other.name
                                          and self.image == other.image
                                          and self.region == other.region
                                          and self.cpu == other.cpu
                                          and self.memory == other.memory
                                          # pylint: disable=line-too-long
                                          and self.nested_virtualization == other.nested_virtualization
                                          and self.dynamic_memory == other.dynamic_memory
                                          and self.networks == other.networks
                                          and self.status == other.status)

    def __repr__(self):
        return f'''
                name: {self.name}
                description: {self.description}
                image: {self.image}
                region: {self.region}
                cpu: {self.cpu}
                memory: {self.memory}
                networks: {self.networks}
                vmm_id: {self.vmm_id}
                status: {self.status}
                region_host: {self.region_host}
                ansible: {self.dados_ansible}
                additional_disks: {self.additional_disks}
                '''

    def to_dict(self):
        dict_objeto = {
            'name': self.name,
            'description': self.description,
            'image': self.image,
            'region': self.region,
            'cpu': self.cpu,
            'memory': self.memory,
            'networks': [network.to_dict() for network in self.networks],
            'vmm_id': self.vmm_id,
            'status': self.status.value,
            'region_host': self.region_host,
            'ansible': [data_ansible.to_dict()
                        for data_ansible in self.dados_ansible.values()]
        }

        if self.to_json_dados_completos:
            dict_objeto['additional_disks'] = [data_disco.to_dict()
                                               for data_disco in self.additional_disks.values()]

        return dict_objeto



"""
General configuration file for the vmm_manager.
"""

# VMM custom fields
FIELD_GROUP = (
    'VMM_MANAGER_AGRUPAMENTO',
    'Resource group (script vmm_manager).'
)

FIELD_ID = (
    'VMM_MANAGER_ID',
    'Resource name in the inventory file (script vmm_manager).'
)

FIELD_IMAGE = (
    'VMM_MANAGER_IMAGEM',
    'Resource template (script vmm_manager).'
)

FIELD_REGION = (
    'VMM_MANAGER_REGIAO',
    'Resource region (script vmm_manager).'
)

FIELD_NETWORK_DEFAULT = (
    'VMM_MANAGER_REDE_PRINCIPAL',
    'Default network (script vmm_manager).'
)



"""
Action entity.
"""
import json

from yamlable import YamlAble, yaml_info

from vmm_manager.infra.command import Command
from vmm_manager.util.config import (FIELD_GROUP, FIELD_ID, FIELD_IMAGE,
                                     FIELD_NETWORK_DEFAULT, FIELD_REGION)


@yaml_info(yaml_tag_ns='scvmm_manager')
class Action(YamlAble):
    ACAO_CRIAR_VM = 'create_vm'
    ACAO_EXCLUIR_VM = 'delete_vm'
    ACAO_EXCLUIR_DISCO_VM = 'delete_vm_disk'

    RESOURCE_IDENTIFIER_NAME = 'vm_name'
    RESOURCE_IDENTIFIER_ID = 'vm_id'

    def __init__(self, command, **kwargs):
        # Validate if the args contain the resource identifier
        if (
            'args' not in kwargs  # passed by Yamlable
            and Action.RESOURCE_IDENTIFIER_NAME not in kwargs
            and Action.RESOURCE_IDENTIFIER_ID not in kwargs
        ):
            raise ValueError(
                f'The args must contain the resource identifier: '
                f'{Action.RESOURCE_IDENTIFIER_NAME} or {Action.RESOURCE_IDENTIFIER_ID}.'
            )

        self.command = command
        self.args = kwargs

        self.__status_execucao = None
        self.__retorno_execucao = None
        self.__status_execucao_job = None

    def executar(self, group, cloud, servidor_acesso, guid):
        cmd = Command(self.command,
                      group=group,
                      field_group=FIELD_GROUP[0],
                      field_id=FIELD_ID[0],
                      field_region=FIELD_REGION[0],
                      cloud=cloud,
                      guid=guid,
                      vmm_server=servidor_acesso.vmm_server
                      )
        cmd.args.update(self.args)

        status, retorno = cmd.executar(servidor_acesso)

        self.__status_execucao = status
        self.__retorno_execucao = json.loads(
            retorno) if self.__status_execucao else retorno

        return self.__status_execucao, self.__retorno_execucao

    def informar_status_execucao_job(self, status):
        self.__status_execucao_job = status

    def is_criacao_vm(self):
        return self.command == Action.ACAO_CRIAR_VM

    def is_bloqueante(self):
        return (self.is_criacao_vm()
                or self.command == Action.ACAO_EXCLUIR_VM
                or self.command == Action.ACAO_EXCLUIR_DISCO_VM)

    def is_same_resource(self, other):
        if (self.RESOURCE_IDENTIFIER_NAME in self.args
                and self.RESOURCE_IDENTIFIER_NAME in other.args):
            return (self.args[self.RESOURCE_IDENTIFIER_NAME]
                    == other.args[self.RESOURCE_IDENTIFIER_NAME])

        if (self.RESOURCE_IDENTIFIER_ID in self.args
                and self.RESOURCE_IDENTIFIER_ID in other.args):
            return self.args[self.RESOURCE_IDENTIFIER_ID] == other.args[self.RESOURCE_IDENTIFIER_ID]

        return False

    def has_cmd_pos_execucao(self):
        return self.is_criacao_vm()

    def was_executada_com_sucesso(self):
        return (self.__status_execucao and
                self.__retorno_execucao.get('Status') == 'OK'
                and (self.__status_execucao_job is None or
                     self.__status_execucao_job))

    def get_resultado_execucao(self):
        if self.__status_execucao is None:
            raise AttributeError('Action not executed.')

        if self.__status_execucao:
            return self.__retorno_execucao

        return {'Msgs': self.__retorno_execucao}

    def get_cmd_pos_execucao(self, group, servidor_acesso):
        if self.is_criacao_vm():
            cmd = Command(
                'create_vm_pos',
                description=f"Add VMM tags for {self.args['vm_name']}",
                vmm_server=servidor_acesso.vmm_server,
                field_group=FIELD_GROUP[0],
                field_id=FIELD_ID[0],
                field_image=FIELD_IMAGE[0],
                field_region=FIELD_REGION[0],
                field_network_default=FIELD_NETWORK_DEFAULT[0],
                group=group
            )
            cmd.args.update(self.args)

            return cmd

        raise AttributeError(
            f'Action "{self.command}" does not have a post execution command.')

    def get_str_impressao_inline(self):
        cmd_args = ', '.join([f'{arg}={value}'
                              for arg, value in self.args.items()])

        return f'{self.command} - [{cmd_args}]'

    def __eq__(self, other):
        return isinstance(other, Action) and (self.command == other.command
                                              and self.args == other.args)

    def __str__(self):
        return f'''
            command: {self.command}
            args: {self.args}
            '''

    def __to_yaml_dict__(self):
        return {
            'command': self.command,
            'args': self.args
        }
